# This code is part of cqlib.
#
# (C) Copyright China Telecom Quantum Group, QuantumCTek Co., Ltd.,
# Center for Excellence in Quantum Information and Quantum Physics 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Quantum circuit"""

from __future__ import annotations

import re
from collections import defaultdict
from typing import Union, Sequence, Optional

from cqlib.circuits import gates
from cqlib.circuits.barrier import Barrier
from cqlib.circuits.instruction import Instruction
from cqlib.circuits.instruction_data import InstructionData
from cqlib.circuits.measure import Measure
from cqlib.circuits.parameter import Parameter
from cqlib.circuits.qubit import Qubit

# Type Alias Definition
Qubits = Union[Qubit, int, Sequence[Union[Qubit, int]]]
IntQubit = Union[Qubit, int]


# pylint: disable=too-many-public-methods
class Circuit:
    """Represents a quantum circuit that can perform various quantum operations."""

    def __init__(self, qubits: Qubits,
                 parameters: Optional[Sequence[Union[Parameter, str]]] = None) -> None:
        """
        Initializes a quantum circuit.

        Args:
            qubits (Qubits): Number of qubits or list of qubits.:
            parameters (list[Parameter | str], optional): List of parameters or parameter names.
        """
        self._qubits = self._initialize_qubits(qubits)
        self._parameters = self._initialize_parameters(parameters)
        self._instruction_sequence: list[InstructionData] = []

    @staticmethod
    def _initialize_parameters(
            parameters: Sequence[Union[Parameter, str]]
    ) -> dict[Union[Parameter, Optional[float]]]:
        """Helper function to initialize parameters."""
        params = {}
        if parameters is None:
            return {}
        for parameter in parameters:
            if isinstance(parameter, str):
                parameter = Parameter(parameter)
            elif not isinstance(parameter, Parameter):
                raise TypeError("Parameters must be of type Parameter or str.")
            if parameter in params:
                raise ValueError(f"Duplicate parameter detected: {parameter}")
            params[parameter] = None
        return params

    @staticmethod
    def _initialize_qubits(qubits: Qubits) -> dict[int, Qubit]:
        """
        Helper function to initialize qubits.

        Args:
            qubits (Qubits): Input qubits specification.

        Returns:
            dict[int, Qubit]: Dictionary of Qubit objects.
        """
        if isinstance(qubits, int):
            if qubits < 0:
                raise ValueError("Number of qubits must be non-negative.")
            return {i: Qubit(i) for i in range(qubits)}
        if isinstance(qubits, Qubit):
            return {qubits.index: qubits}
        if isinstance(qubits, (list, tuple)):
            qs = {}
            for qubit in qubits:
                if isinstance(qubit, int):
                    qubit = Qubit(qubit)
                elif not isinstance(qubit, Qubit):
                    raise TypeError("Qubit must be an int or QuBit instance.")
                if qubit.index in qs:
                    raise ValueError(f"Duplicate qubit detected: {qubit}")
                qs[qubit.index] = qubit
            return qs
        raise ValueError("Invalid qubits input. Expected int, QuBit, or list/tuple of these.")

    @property
    def qubits(self) -> list[Qubit]:
        """circuit qubits"""
        return list(self._qubits.values())

    @property
    def num_qubits(self) -> int:
        """Return number of qubits."""
        return len(self._qubits)

    def qubits_path(self) -> dict[Qubit, list[InstructionData]]:
        """
        Constructs a path of operations for each qubit.
        """
        path = defaultdict(list)
        max_depths = defaultdict(int)
        for op in self._instruction_sequence:
            instruction = op.instruction
            qubits = op.qubits
            if instruction.num_qubits != len(qubits):
                raise ValueError("Instruction qubit count does not match provided qubits.")
            current_max_depth = max(max_depths[qubit] for qubit in qubits)
            for qubit in qubits:
                ops = [None] * (current_max_depth - max_depths[qubit]) + [op]
                path[qubit].extend(ops)
                max_depths[qubit] = current_max_depth + 1
        return path

    def depth(self) -> int:
        """Max length of critical path"""
        path = self.qubits_path()
        return max(len(ops) for _, ops in path.items())

    def add_qubit(self, qubit: IntQubit):
        """
        Adds a qubit to the circuit, ensuring it does not already exist.

        Args:
            qubit (IntQubit): The qubit to add, specified as an integer index or a Qubit object.
        """
        if isinstance(qubit, int):
            qubit = Qubit(qubit)
        elif not isinstance(qubit, Qubit):
            raise TypeError("Qubit must be an int or QuBit instance.")
        if qubit.index in self._qubits:
            raise ValueError("Qubit already exists in the circuit.")
        self._qubits[qubit.index] = qubit

    @property
    def parameters(self) -> list[Parameter]:
        """
        Retrieves a list of all parameters currently in the quantum circuit.

        Returns:
            list[Parameter]: A list containing the parameters used in the circuit.
        """
        return list(self._parameters.keys())

    def add_parameter(self, parameter: Union[Parameter, str]):
        """
        Adds a new parameter to the quantum circuit.
        If the parameter already exists, it raises an error.

        Args:
            parameter (Parameter | str): The parameter to add, can be an instance of
                Parameter or a string name. If a string is provided, a new Parameter
                instance will be created from it.
        """
        if isinstance(parameter, str):
            parameter = Parameter(parameter)
        elif not isinstance(parameter, Parameter):
            raise TypeError("Parameters must be of type Parameter or str.")
        if parameter in self._parameters:
            raise ValueError("Parameter already exists in the circuit.")
        self._parameters[parameter] = None

    @property
    def instruction_sequence(self) -> list[InstructionData]:
        """
        Accesses the list of all instruction data objects stored in the circuit.

        Returns:
            list[InstructionData]: The sequence of instructions added to the circuit,
                each represented by an InstructionData object.
        """
        return self._instruction_sequence

    def insert(self, instruction: Instruction, qubits: Qubits, index: int = None):
        """
        Insert an instruction into the circuit at a specified index.

        Args:
            instruction (Instruction): The instruction to be inserted.
            qubits (Qubits): The qubit(s) to which the instruction is applied.
            index (int, optional): The index at which to insert the instruction.
                Defaults to None, which appends at the end.
        """
        instruction, qubits = self._prepare_instruction(instruction, qubits)
        if index is None:
            index = len(self._instruction_sequence)
        elif not isinstance(index, int):
            raise TypeError("Index must be an integer or None.")
        elif index < 0 or index > len(self._instruction_sequence):
            raise ValueError(f"Index {index} out of bounds for instructions list of length "
                             f"{len(self._instruction_sequence)}.")

        self._instruction_sequence.insert(
            index,
            InstructionData(instruction=instruction, qubits=qubits)
        )

    def append(self, instruction: Instruction, qubits: Qubits):
        """
        append instruction to circuit

        Args:
            instruction (Instruction): The instruction to be appended.
            qubits (Qubits): The qubit(s) to which the instruction is applied.
        """
        instruction, qubits = self._prepare_instruction(instruction, qubits)
        self._instruction_sequence.append(InstructionData(instruction=instruction, qubits=qubits))

    def append_instruction_data(self, instruction_data: InstructionData):
        """
        append instruction_data to circuit

        Args:
            instruction_data (InstructionData): The instruction_data to be appended.
        """
        self._instruction_sequence.append(instruction_data)

    def _prepare_instruction(self, instruction: Instruction, qubits: Qubits):
        """
        append instruction to circuit

        Args:
            instruction (Instruction): The instruction to be appended.
            qubits (Qubits): The qubit(s) to which the instruction is applied.
        """
        qs = self._normalize_qubits(qubits)
        if instruction.num_qubits != len(qs):
            raise ValueError(f"Instruction requires {instruction.num_qubits} qubits,"
                             f" but {len(qubits)} were given.")
        self._update_instruction_params(instruction)
        return instruction, qs

    def _normalize_qubits(self, qubits: Qubits) -> list[Qubit]:
        """ Normalize and validate qubits input. """
        if not isinstance(qubits, (list, tuple)):
            qubits = [qubits]

        qs = []
        for q in qubits:
            if isinstance(q, int):
                q = Qubit(q)
            if not isinstance(q, Qubit):
                raise TypeError("Qubit must be an instance of QuBit or an integer.")
            if q.index not in self._qubits:
                raise ValueError(f"{q} not found in circuit.")
            qs.append(q)
        return qs

    def _update_instruction_params(self, instruction: Instruction):
        """ Validate and set parameters for the instruction. """
        if not instruction.params:
            return
        params = []
        for param in instruction.params:
            if isinstance(param, str):
                param = Parameter(param)
            if not isinstance(param, (Parameter, float, int)):
                raise TypeError("Parameter must be a float, int, str, or an instance of Parameter.")
            if isinstance(param, Parameter):
                for p in param.base_params:
                    if p not in self._parameters:
                        raise ValueError(f"Parameter {p} not found in circuit.")
            params.append(param)
        instruction.params = params

    def h(self, qubit: IntQubit):
        """
        Applies the Hadamard gate (H) to a specified qubit.

        Args:
            qubit (IntQubit): The qubit to apply the Hadamard gate.
        """
        self.append(gates.H(), [qubit])

    def i(self, qubit: IntQubit, t: int):
        """
        Applies the `I` gate to a specified qubit.

        Args:
            qubit (IntQubit): The qubit to apply the Hadamard gate.
            t (int): The duration the gate acts, in integer units of 0.5 ns.
        """
        self.append(gates.I(t), [qubit])

    def rx(self, qubit: IntQubit, theta: Union[float, Parameter]):
        """
        Applies the RX gate (rotation around the X-axis) to a specified
            qubit with a given angle.

        Args:
            qubit (IntQubit): The qubit to apply the RX gate.
            theta (Union[float, Parameter]): The rotation angle in radians or as a symbolic parameter.
        """
        self.append(gates.RX(theta), [qubit])

    def rxy(self, qubit: IntQubit, phi: Union[float , Parameter], theta: Union[float , Parameter]):
        """
        Applies the RX gate (rotation around the X-axis) to a specified
            qubit with a given angle.

        Args:
            qubit (IntQubit): The qubit to apply the RX gate.
            phi (Union[float, Parameter]): The rotation angle in radians or as a symbolic parameter.
            theta (Union[float, Parameter]): The rotation angle in radians or as a symbolic parameter.
        """
        self.append(gates.RXY(phi, theta), [qubit])

    def crx(self, control_qubit: IntQubit, target_qubit: IntQubit, theta: Union[float, Parameter]):
        """
        Applies the controlled-RX gate (controlled rotation around the X-axis) between two qubits.

        Args:
            control_qubit (IntQubit): The control qubit.
            target_qubit (IntQubit): The target qubit where the rotation is applied.
            theta (Union[float, Parameter]): The rotation angle in radians or as a symbolic parameter.
        """
        self.append(gates.CRX(theta), [control_qubit, target_qubit])

    def ry(self, qubit: IntQubit, theta: Union[float, Parameter]):
        """
         Applies the RY gate (rotation around the Y-axis) to a specified qubit with a given angle.

         Args:
             qubit (IntQubit): The qubit to apply the RY gate.
             theta (Union[float, Parameter]): The rotation angle in radians or as a symbolic parameter.
         """
        self.append(gates.RY(theta), [qubit])

    def cry(self, control_qubit: IntQubit, target_qubit: IntQubit, theta: Union[float, Parameter]):
        """
        Applies the controlled-RY gate (controlled rotation around the Y-axis) between two qubits.

        Args:
            control_qubit (IntQubit): The control qubit.
            target_qubit (IntQubit): The target qubit where the rotation is applied.
            theta (Union[float, Parameter]): The rotation angle in radians or as a symbolic parameter.
        """
        self.append(gates.CRY(theta), [control_qubit, target_qubit])

    def rz(self, qubit: IntQubit, theta: Union[float, Parameter]):
        """
        Applies the RZ gate (rotation around the Z-axis) to a specified qubit with a given angle.

        Args:
            qubit (IntQubit): The qubit to apply the RZ gate.
            theta (Union[float, Parameter]): The rotation angle in radians or as a symbolic parameter.
        """
        self.append(gates.RZ(theta), [qubit])

    def crz(self, control_qubit: IntQubit, target_qubit: IntQubit, theta: Union[float, Parameter]):
        """
        Applies the controlled-RZ gate (controlled rotation around the Z-axis) between two qubits.

        Args:
            control_qubit (IntQubit): The control qubit.
            target_qubit (IntQubit): The target qubit where the rotation is applied.
            theta (Union[float, Parameter]): The rotation angle in radians or as a symbolic parameter.
        """
        self.append(gates.CRZ(theta), [control_qubit, target_qubit])

    def s(self, qubit: IntQubit):
        """
        Applies the S gate (phase gate) to a specified qubit.

        Args:
            qubit (IntQubit): The qubit to apply the S gate.
        """
        self.append(gates.S(), [qubit])

    def sd(self, qubit: IntQubit):
        """
        Applies the S-dagger gate (inverse S gate) to a specified qubit.

        Args:
            qubit (IntQubit): The qubit to apply the S-dagger gate.
        """
        self.append(gates.SD(), [qubit])

    def swap(self, qubit1: IntQubit, qubit2: IntQubit):
        """
        Applies the SWAP gate to exchange the states of two specified qubits.

        Args:
            qubit1 (IntQubit): The first qubit.
            qubit2 (IntQubit): The second qubit.
        """
        self.append(gates.SWAP(), [qubit1, qubit2])

    def t(self, qubit: IntQubit):
        """
        Applies the T gate (π/8 gate) to a specified qubit.

        Args:
            qubit (IntQubit): The qubit to apply the T gate.
        """
        self.append(gates.T(), [qubit])

    def td(self, qubit: IntQubit):
        """
        Applies the TD gate (-π/8 gate) to a specified qubit.

        Args:
            qubit (IntQubit): The qubit to apply the T gate.
        """
        self.append(gates.TD(), [qubit])

    def x(self, qubit: IntQubit):
        """
        Applies the X gate (Pauli X, flip gate) to a specified qubit.

        Args:
            qubit (IntQubit): The qubit to apply the X gate.
        """
        self.append(gates.X(), [qubit])

    def cx(self, control_qubit: IntQubit, target_qubit: IntQubit):
        """
        Applies the CX gate (controlled-X, CNOT) between a control qubit and a target qubit.

        Args:
            control_qubit (IntQubit): The control qubit.
            target_qubit (IntQubit): The target qubit where the NOT operation is applied
             conditionally.
        """
        self.append(gates.CX(), [control_qubit, target_qubit])

    def ccx(self, control_qubit_1: IntQubit, control_qubit_2: IntQubit, target_qubit: IntQubit):
        """
        Applies the CCX gate (Toffoli gate, controlled-controlled-X) using two control qubits
        and one target qubit.

        Args:
            control_qubit_1 (IntQubit): The first control qubit.
            control_qubit_2 (IntQubit): The second control qubit.
            target_qubit (IntQubit): The target qubit where the X operation is applied
                conditionally.
        """
        self.append(gates.CCX(), [control_qubit_1, control_qubit_2, target_qubit])

    def x2p(self, qubit: IntQubit):
        """
        Applies the X2P gate to a specified qubit.

        Args:
            qubit (IntQubit): The qubit to apply the X2P gate.
        """
        self.append(gates.X2P(), [qubit])

    def x2m(self, qubit: IntQubit):
        """
         Applies the X2M gate to a specified qubit.

         Args:
             qubit (IntQubit): The qubit to apply the X2M gate.
         """
        self.append(gates.X2M(), [qubit])

    def xy(self, qubit: IntQubit, theta: Union[float, Parameter]):
        """
        Applies the XY gate (generalized rotation around an axis in the XY plane)
        to a specified qubit.

        Args:
            qubit (IntQubit): The qubit to apply the XY gate.
            theta (Union[float, Parameter]): The rotation angle around the XY axis.
        """
        self.append(gates.XY(theta), [qubit])

    def xy2p(self, qubit: IntQubit, theta: Union[float, Parameter]):
        """
        Applies the XY2P gate to a specified qubit.

        Args:
            qubit (IntQubit): The qubit to apply the XY2P gate.
            theta (Union[float, Parameter]): The rotation angle for the root operation.
        """
        self.append(gates.XY2P(theta), [qubit])

    def xy2m(self, qubit: IntQubit, theta: Union[float, Parameter]):
        """
        Applies the XY2M gate to a specified qubit.

        Args:
            qubit (IntQubit): The qubit to apply the XY2M gate.
            theta (Union[float, Parameter]): The rotation angle for the root operation.
        """
        self.append(gates.XY2M(theta), [qubit])

    def y(self, qubit: IntQubit):
        """
        Applies the Y gate (Pauli Y) to a specified qubit.

        Args:
            qubit (IntQubit): The qubit to apply the Y gate.
        """
        self.append(gates.Y(), [qubit])

    def y2p(self, qubit: IntQubit):
        """
        Applies the Y2P gate to a specified qubit.

        Args:
            qubit (IntQubit): The qubit to apply the Y2P gate.
        """
        self.append(gates.Y2P(), [qubit])

    def y2m(self, qubit: IntQubit):
        """
        Applies the Y2M gate to a specified qubit.

        Args:
            qubit (IntQubit): The qubit to apply the Y2M gate.
        """
        self.append(gates.Y2M(), [qubit])

    def z(self, qubit: IntQubit):
        """
       Applies the Z gate (Pauli Z) to a specified qubit.

       Args:
           qubit (IntQubit): The qubit to apply the Z gate.
       """
        self.append(gates.Z(), [qubit])

    def cz(self, control_qubit: IntQubit, target_qubit: IntQubit):
        """
        Applies the CZ gate (controlled-Z) between a control qubit and a target qubit.

        Args:
            control_qubit (IntQubit): The control qubit.
            target_qubit (IntQubit): The target qubit where the Z operation is applied
             conditionally.
        """
        self.append(gates.CZ(), [control_qubit, target_qubit])

    def barrier(self, *qubits: IntQubit):
        """
        Inserts a barrier into the circuit that affects all specified qubits.
        A barrier ensures that no optimizations or rearrangements are done across
        this point in the circuit, maintaining the order of operations.

        Args:
            qubits (IntQubit): The qubits to which the barrier is applied.
        """
        self.append(Barrier(len(qubits)), list(qubits))

    def measure(self, qubits: Qubits):
        """
        Measures the specified qubits and collapses their quantum state into classical bits.

        Args:
            qubits (Qubits): The qubits to measure, can be a single qubit or a list of qubits.
        """
        if isinstance(qubits, (Qubit, int)):
            qubits = [qubits]

        for q in qubits:
            self.append(Measure(), [q])

    def measure_all(self):
        """
        Measures all qubits in the circuit that have not yet been measured.
        """
        measured_qubits = {ins[1][0] for ins in self._instruction_sequence
                           if isinstance(ins[0], Measure)}
        unmeasured_qubits = [q for index, q in self._qubits.items() if q not in measured_qubits]
        for qubit in unmeasured_qubits:
            self.append(Measure(), qubit)

    def set_parameter_value(self, values: Optional[dict[Union[str, Parameter], Union[float , int]]] = None, **kwargs):
        """
        Sets the values of specified parameters in the quantum circuit.
        This method allows parameters to be set using a dictionary or
        directly via keyword arguments.

        Args:
            values (dict[str | Parameter, float | int], optional): A
                dictionary where keys are parameters (either as Parameter
                objects or their string identifiers) and values are the
                numerical values (float or int) to set. If a parameter is
                specified as a string, it must be a valid identifier and
                already exist in the circuit's parameter list.

            **kwargs: Additional parameters and their values provided as
                keyword arguments. This is useful for directly setting values
                when calling the method.

        Example:
            >>> from cqlib.circuits import Circuit
            >>> circuit = Circuit(2, ['theta', 'phi'])
            >>> circuit.set_parameter_value({'theta': 3.14159})
            >>> circuit.set_parameter_value(theta=0.5, phi=0.25)

        This method updates the internal dictionary of parameters to
        reflect the new values, allowing these parameters to be used
        with their new values in subsequent operations within the circuit.
        """
        if values is None:
            values = {}
        values.update(kwargs)
        for param, value in values.items():
            if not isinstance(value, (float, int)):
                raise ValueError("Parameter value must be a float or int.")

            if isinstance(param, Parameter):
                if param not in self._parameters:
                    raise KeyError(f"Parameter {param} not found.")
            elif isinstance(param, str):
                if not param.isidentifier():
                    raise ValueError(f"Invalid parameter identifier: {param}")
                params = [p for p in self._parameters if p == param]
                if not params:
                    raise ValueError(f"Parameter {param} not found in circuit.")
                param = params[0]
            else:
                raise TypeError(f"Parameter identifier must be a string or Parameter,"
                                f" got {type(param).__name__}.")
            self._parameters[param] = value

    @property
    def qcis(self) -> str:
        """
        Generates a qcis string of all instructions in the circuit.
        """
        return self._export_circuit_str(self._instruction_sequence, True, self._parameters)

    def export_circuit_str(self, qcis_compliant: bool = False):
        """
        Exports the circuit as a string format, with an option to make the output QCIS-compliant.

        Args:
            qcis_compliant (bool): If True, the output will conform to QCIS standards;
             if False, it will retain the original format with composite gates.
        """
        params = self._parameters
        return self._export_circuit_str(self._instruction_sequence, qcis_compliant, params)

    @classmethod
    def _export_circuit_str(
            cls,
            instructions: list[InstructionData],
            qcis_compliant: bool = False,
            params=None
    ):
        """
        Exports the circuit as a string format, with an option to make the output QCIS-compliant.

        Args:
            instructions: A list of InstructionData.
            qcis_compliant (bool): If True, the output will conform to QCIS standards;
             if False, it will retain the original format with composite gates.
        """
        if params is None:
            params = {}
        ops = []
        for instruction_data in instructions:
            instruction = instruction_data.instruction
            qubits = instruction_data.qubits

            if not isinstance(instruction, Instruction) or instruction.num_qubits != len(qubits):
                raise ValueError(f"Instruction {instruction.name} configuration error"
                                 f" with qubits {qubits}")
            if qcis_compliant and not instruction.is_supported_by_qcis:
                ops.append(cls._export_circuit_str(instruction.to_qcis(qubits), qcis_compliant))
                continue

            line = [instruction.name] + [str(qubit) for qubit in qubits]
            for param in instruction.params:
                if isinstance(param, Parameter):
                    value = param.value(params)
                    if value.is_Float:
                        line.append(str(value).rstrip('0').rstrip('.'))
                    else:
                        line.append(str(value))
                else:
                    line.append(str(param))
            ops.append(' '.join(line))
        return '\n'.join(ops)

    @staticmethod
    def load(qcis: str) -> Circuit:
        """
            Loads a quantum circuit from a formatted string.

            Args:
                qcis (str): A string containing quantum circuit instructions, where each line
                            represents a circuit operation. The format for each line is
                            "GATE QUBITS [PARAMETERS]", e.g., "H Q0 Q1", "CX Q0 Q1", "RZ Q0 0.5".

            Returns:
                Circuit: A quantum circuit object constructed based on the input string.

            Raises:
                ValueError: If the input string is improperly formatted or contains
                    unknown gate operations.

            Example:
                >>> circuit_description = "H Q0\\nCX Q0 Q1\\nM Q0"
                >>> c = Circuit.load(circuit_description)
                >>> print(c)
                Circuit with 3 instructions
            """
        # from cqlib.circuits import gates, Measure, Barrier

        circuit = Circuit(0)
        line_pattern = re.compile(r'^([A-Z][A-Z0-9]*)\s+((?:Q[0-9]+\s*)+)((?:-?\d*\.?\d+\s*)*)$')

        for line in qcis.split('\n'):
            line = line.strip().upper()
            if not line:
                continue
            match = line_pattern.match(line)
            if not match:
                raise ValueError(f'Invalid line format: {line}')

            gate, qubits_str, params_str = match.groups()
            qubits = [circuit._qubits.setdefault(int(q[1:]), Qubit(int(q[1:])))
                      for q in qubits_str.split()]
            params = [float(p) for p in params_str.split()] if params_str else []

            if gate == 'M':
                for qubit in qubits:
                    circuit.append_instruction_data(
                        InstructionData(instruction=Measure(), qubits=[qubit]))
                continue
            if gate == 'B':
                circuit.append_instruction_data(
                    InstructionData(instruction=Barrier(len(qubits)), qubits=qubits))
                continue

            if not hasattr(gates, gate):
                raise ValueError(f'Unknown gate: {gate}')
            data = InstructionData(instruction=getattr(gates, gate)(*params, label=None),
                                   qubits=qubits)
            circuit.append_instruction_data(data)

        return circuit
