# This code is part of cqlib.
#
# Copyright (C) 2024 China Telecom Quantum Group, QuantumCTek Co., Ltd.,
# Center for Excellence in Quantum Information and Quantum Physics.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Defines the Pauli Y gate and its variants: Y2P (rotations around the y-axis by pi/2)
and Y2M (rotations around the y-axis by -pi/2).

Classes:
    Y: Pauli Y gate, acts as a quantum bit flip and phase shift gate.
    Y2P: Performs a rotation around the y-axis by pi/2.
    Y2M: Performs a rotation around the y-axis by -pi/2.
"""
from typing import Optional
import numpy as np

from cqlib.circuits.gates.gate import Gate
from cqlib.circuits.utils import sqrt2_inv


class Y(Gate):
    """
    Pauli Y gate.

    This gate acts as both a bit flip and a phase shift on a qubit.
    It is represented by the matrix:
    [[ 0, -i ],
    [ i, 0 ]]
    """

    def __init__(self, label: Optional[str] = None):
        """
        Initialize the Pauli Y gate.

        Args:
            label (str | None, optional): An optional label for the Y gate. Defaults to None.
        """
        super().__init__('Y', 1, [], label=label)

    def __array__(self, dtype: np.dtype = np.complex128):
        """
        The numpy matrix of the Y gate.

        Args:
            dtype (optional): The data type of the matrix elements. Default is np.complex128.

        Returns:
            numpy.ndarray: A 2x2 matrix with complex entries representing the Y gate.
        """
        return np.array([[0, -1j], [1j, 0]], dtype=dtype)


class Y2P(Gate):
    """
    Y2P gate performs a rotation around the y-axis of the Bloch sphere by pi/2.

    This rotation is useful for creating specific quantum states or for quantum state manipulation.
    It is represented by the matrix:
    sqrt(1/2) * [[ 1, -1 ], [ 1, 1 ]]
    """

    def __init__(self, label: Optional[str] = None):
        """
        Initialize the Y2P gate.

        Args:
            label (str | None, optional): An optional label for the Y2P gate. Defaults to None.
        """
        super().__init__('Y2P', 1, [], label=label)

    def __array__(self, dtype: np.dtype = np.complex128):
        """
        The numpy matrix of the Y2P gate.

        Args:
            dtype (optional): The data type of the matrix elements. Default is np.complex128.

        Returns:
            numpy.ndarray: A 2x2 matrix with complex entries representing the Y2P gate.
        """
        return sqrt2_inv * np.array([[1, -1], [1, 1]], dtype=dtype)


class Y2M(Gate):
    """
    Y2M gate performs a rotation around the Y-axis of the Bloch sphere by -pi/2.

    This rotation is useful for undoing specific quantum operations or preparing quantum states.
    It is represented by the matrix:
    sqrt(1/2) * [[ 1, 1 ], [ -1, 1 ]]
    """

    def __init__(self, label: Optional[str] = None):
        """
        Initialize the Y2M gate.

        Args:
            label (str | None, optional): An optional label for the Y2M gate. Defaults to None.
        """
        super().__init__('Y2M', 1, [], label=label)

    def __array__(self, dtype: np.dtype = np.complex128):
        """
        The numpy matrix of the Y2M gate.

        Args:
            dtype (optional): The data type of the matrix elements. Default is np.complex128.

        Returns:
            numpy.ndarray: A 2x2 matrix with complex entries representing the Y2M gate.
        """
        return sqrt2_inv * np.array([[1, 1], [-1, 1]], dtype=dtype)
